/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.test.ha.framework.server;

import java.util.EventObject;
import java.util.List;

import junit.framework.TestCase;

import org.easymock.EasyMock;
import org.jboss.ha.framework.interfaces.DistributedReplicantManager;
import org.jboss.ha.framework.interfaces.EventListener;
import org.jboss.ha.framework.interfaces.HAPartition;
import org.jboss.ha.framework.server.HAServiceEvent;
import org.jboss.ha.framework.server.HAServiceEventFactory;
import org.jboss.ha.framework.server.HAServiceImpl;
import org.jboss.ha.framework.server.HAServiceRpcHandler;

/**
 * Unit test for {@link HAServiceImpl}.
 * @author Paul Ferraro
 */
public class HAServiceTestCase extends TestCase
{
   protected static final String SERVICE_HA_NAME = "test";
   
   protected final HAPartition partition = EasyMock.createStrictMock(HAPartition.class);
   protected final DistributedReplicantManager drm = EasyMock.createStrictMock(DistributedReplicantManager.class);
   
   private final HAServiceImpl<HAServiceEvent> customService = new HAServiceImpl<HAServiceEvent>(new HAServiceEventFactory())
   {
      @Override
      protected void partitionTopologyChanged(List<?> newReplicants, int newReplicantsViewId, boolean merge)
      {
         newReplicants.clear();
      }
   };

   private HAServiceImpl<HAServiceEvent> service;
   
   /**
    * @see junit.framework.TestCase#setUp()
    */
   @Override
   protected void setUp() throws Exception
   {
      this.service = this.getHAService();
      
      this.service.setHAPartition(this.partition);
      this.service.setServiceHAName(SERVICE_HA_NAME);
   }
   
   protected Object getRpcHandler()
   {
      return EasyMock.isA(HAServiceRpcHandler.class);
   }
   
   protected HAServiceImpl<HAServiceEvent> getHAService()
   {
      return this.customService;
   }
   
   public void testGetServiceKey()
   {
      String key = this.service.getHAServiceKey();
      
      assertSame(SERVICE_HA_NAME, key);
   }
   
   public void testStart() throws Exception
   {
      // Test isRegisterThreadContextClassLoader() = default value
      this.partition.registerRPCHandler(EasyMock.same(SERVICE_HA_NAME), this.getRpcHandler());
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      this.drm.registerListener(SERVICE_HA_NAME, this.service);
      this.drm.add(SERVICE_HA_NAME, "");
      
      EasyMock.replay(this.partition, this.drm);
      
      this.service.start();
      
      EasyMock.verify(this.partition, this.drm);
      EasyMock.reset(this.partition, this.drm);
      
      // Test isRegisterThreadContextClassLoader() = true
      this.service.setRegisterThreadContextClassLoader(true);
      
      this.partition.registerRPCHandler(EasyMock.same(SERVICE_HA_NAME), this.getRpcHandler(), EasyMock.same(Thread.currentThread().getContextClassLoader()));
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      this.drm.registerListener(SERVICE_HA_NAME, this.service);
      this.drm.add(SERVICE_HA_NAME, "");
      
      EasyMock.replay(this.partition, this.drm);
      
      this.service.start();
      
      EasyMock.verify(this.partition, this.drm);
      EasyMock.reset(this.partition, this.drm);
   }
   
   public void testStop() throws Exception
   {
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      this.drm.remove(SERVICE_HA_NAME);
      this.drm.unregisterListener(SERVICE_HA_NAME, this.service);

      this.partition.unregisterRPCHandler(EasyMock.same(SERVICE_HA_NAME), this.getRpcHandler());
      
      EasyMock.replay(this.partition, this.drm);
      
      this.service.stop();
      
      EasyMock.verify(this.partition, this.drm);
      EasyMock.reset(this.partition, this.drm);
   }
   
   public void testHAPartition()
   {
      HAPartition partition = this.service.getHAPartition();
      
      assertSame(this.partition, partition);
      
      HAPartition partition2 = EasyMock.createMock(HAPartition.class);
      
      this.service.setHAPartition(partition2);
      
      partition = this.service.getHAPartition();

      assertSame(partition2, partition);
   }
   
   public void testHAServiceName()
   {
      String name = this.service.getServiceHAName();
      
      assertSame(SERVICE_HA_NAME, name);
      
      final String newName = "new";
      
      this.service.setServiceHAName(newName);
      
      name = this.service.getServiceHAName();

      assertSame(newName, name);
   }
   
   public void testHandleEvent() throws Exception
   {
      HAServiceEvent event = new HAServiceEvent("", "");
      
      this.partition.callAsynchMethodOnCluster(EasyMock.same(SERVICE_HA_NAME), EasyMock.eq("handleEvent"), EasyMock.aryEq(new Object[] { event }), EasyMock.aryEq(new Class[] { EventObject.class }), EasyMock.eq(true));
      
      EasyMock.replay(this.partition);
      
      this.service.handleEvent(event);
      
      EasyMock.verify(this.partition);
      EasyMock.reset(this.partition);
   }
   
   public void testRegisterThreadContextClassLoader()
   {
      boolean register = this.service.isRegisterThreadContextClassLoader();
      
      assertFalse(register);
      
      this.service.setRegisterThreadContextClassLoader(true);
      
      register = this.service.isRegisterThreadContextClassLoader();

      assertTrue(register);
   }
   
   @SuppressWarnings("unchecked")
   public void testEventFacility()
   {
      // Test add and notify
      HAServiceEvent event = new HAServiceEvent("", "");
      EventListener<HAServiceEvent> listener1 = EasyMock.createStrictMock(EventListener.class);
      EventListener<HAServiceEvent> listener2 = EasyMock.createStrictMock(EventListener.class);
      
      try
      {
         // Verify that next listener is notified even if 1st throws exception
         listener1.handleEvent(event);
         EasyMock.expectLastCall().andThrow(new Exception());
         
         listener2.handleEvent(event);
      }
      catch (Exception e)
      {
         fail(e.toString());
      }
      
      EasyMock.replay(listener1, listener2);
      
      this.service.addEventListener(listener1);
      this.service.addEventListener(listener2);
      this.service.notifyListeners(event);
      
      EasyMock.verify(listener1, listener2);
      EasyMock.reset(listener1, listener2);
      
      // Test remove
      try
      {
         // Verify that next listener is notified even if 1st throws exception
         listener1.handleEvent(event);
      }
      catch (Exception e)
      {
         fail(e.toString());
      }
      
      EasyMock.replay(listener1, listener2);
      
      this.service.removeEventListener(listener2);
      this.service.notifyListeners(event);
      
      EasyMock.verify(listener1, listener2);
      EasyMock.reset(listener1, listener2);
   }
   
   public void testReplicantsChanged() throws Exception
   {
      List<?> list = EasyMock.createStrictMock(List.class);
      
      EasyMock.replay(list);
      
      this.service.replicantsChanged("", list, 100, false);
      
      EasyMock.verify(list);
      EasyMock.reset(list);
      
      list.clear();
      
      EasyMock.replay(list);
      
      this.service.replicantsChanged(SERVICE_HA_NAME, list, 101, false);
      
      EasyMock.verify(list);
      EasyMock.reset(list);
   }
}
