/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.test.ha.framework.server;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.easymock.Capture;
import org.easymock.EasyMock;
import org.jboss.ha.framework.interfaces.ClusterNode;
import org.jboss.ha.framework.interfaces.DistributedReplicantManager;
import org.jboss.ha.framework.interfaces.HASingletonElectionPolicy;
import org.jboss.ha.framework.interfaces.HASingletonLifecycle;
import org.jboss.ha.framework.server.HAServiceEvent;
import org.jboss.ha.framework.server.HAServiceEventFactory;
import org.jboss.ha.framework.server.HAServiceImpl;
import org.jboss.ha.framework.server.HASingletonImpl;
import org.jboss.ha.framework.server.HASingletonRpcHandler;

/**
 * Unit test for {@link HASingletonImpl}.
 * @author Paul Ferraro
 */
public class HASingletonTestCase extends HAServiceTestCase
{
   private static final String SERVICE_HA_NAME = "test";
   
   private final HASingletonElectionPolicy electionPolicy = EasyMock.createStrictMock(HASingletonElectionPolicy.class);
   final HASingletonLifecycle lifecycle = EasyMock.createStrictMock(HASingletonLifecycle.class);
   
   private final HASingletonImpl<HAServiceEvent> customSingleton = new HASingletonImpl<HAServiceEvent>(new HAServiceEventFactory())
   {
      @Override
      public void startSingleton()
      {
         HASingletonTestCase.this.lifecycle.startSingleton();
      }
      
      @Override
      public void stopSingleton()
      {
         HASingletonTestCase.this.lifecycle.stopSingleton();
      }
   };

   private HASingletonImpl<HAServiceEvent> singleton;
   
   /**
    * @see org.jboss.test.ha.framework.server.HAServiceTestCase#setUp()
    */
   @Override
   protected void setUp() throws Exception
   {
      super.setUp();
      
      this.singleton = this.getHASingleton();
   }

   protected HASingletonImpl<HAServiceEvent> getHASingleton()
   {
      return this.customSingleton;
   }
   
   @Override
   protected HAServiceImpl<HAServiceEvent> getHAService()
   {
      return this.customSingleton;
   }

   @Override
   protected Object getRpcHandler()
   {
      return EasyMock.isA(HASingletonRpcHandler.class);
   }
   
   @Override
   public void testStart() throws Exception
   {
      Capture<DistributedReplicantManager.ReplicantListener> capturedListener1 = new Capture<DistributedReplicantManager.ReplicantListener>();
      Capture<DistributedReplicantManager.ReplicantListener> capturedListener2 = new Capture<DistributedReplicantManager.ReplicantListener>();
      
      // Test isRegisterThreadContextClassLoader() = default value
      this.partition.registerRPCHandler(EasyMock.same(SERVICE_HA_NAME), this.getRpcHandler());
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      
      this.drm.registerListener(EasyMock.same(SERVICE_HA_NAME), EasyMock.capture(capturedListener1));
      this.drm.add(SERVICE_HA_NAME, "");
      this.drm.registerListener(SERVICE_HA_NAME, this.singleton);
      this.drm.unregisterListener(EasyMock.same(SERVICE_HA_NAME), EasyMock.capture(capturedListener2));
      
      EasyMock.replay(this.partition, this.drm);
      
      this.singleton.start();
      
      EasyMock.verify(this.partition, this.drm);
      
      assertSame(capturedListener1.getValue(), capturedListener2.getValue());
      
      EasyMock.reset(this.partition, this.drm);
      
      // Test isRegisterThreadContextClassLoader() = true
      this.singleton.setRegisterThreadContextClassLoader(true);
      
      this.partition.registerRPCHandler(EasyMock.same(SERVICE_HA_NAME), this.getRpcHandler(), EasyMock.same(Thread.currentThread().getContextClassLoader()));
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      
      this.drm.registerListener(EasyMock.same(SERVICE_HA_NAME), EasyMock.capture(capturedListener1));
      this.drm.add(SERVICE_HA_NAME, "");
      this.drm.registerListener(SERVICE_HA_NAME, this.singleton);
      this.drm.unregisterListener(EasyMock.same(SERVICE_HA_NAME), EasyMock.capture(capturedListener2));
      
      EasyMock.replay(this.partition, this.drm);
      
      this.singleton.start();
      
      EasyMock.verify(this.partition, this.drm);
      
      assertSame(capturedListener1.getValue(), capturedListener2.getValue());
      
      EasyMock.reset(this.partition, this.drm);
   }
   
   public void testRestartOnMerge()
   {
      boolean restart = this.singleton.getRestartOnMerge();
      
      assertTrue(restart);
      
      this.singleton.setRestartOnMerge(false);
      
      restart = this.singleton.getRestartOnMerge();

      assertFalse(restart);
   }
   
   public void testElectionPolicy()
   {
      HASingletonElectionPolicy policy = this.singleton.getElectionPolicy();
      
      assertNull(policy);
      
      this.singleton.setElectionPolicy(this.electionPolicy);
      
      policy = this.singleton.getElectionPolicy();

      assertSame(this.electionPolicy, policy);
   }
   
   public void testIsMaster()
   {
      boolean master = this.singleton.isMasterNode();
      
      assertFalse(master);
   }
   
   public void testStopOldMaster() throws Exception
   {
      this.partition.callAsynchMethodOnCluster(EasyMock.same(SERVICE_HA_NAME), EasyMock.eq("stopOldMaster"), EasyMock.aryEq(new Object[0]), EasyMock.aryEq(new Class[0]), EasyMock.eq(true));
      
      EasyMock.replay(this.partition);
      
      this.singleton.stopOldMaster();
      
      EasyMock.verify(this.partition);
      EasyMock.reset(this.partition);
   }
   
   @Override
   public void testReplicantsChanged() throws Exception
   {
      List<?> list = Collections.emptyList();
      
      // Test non-matching service name
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged("", list, 100, false);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertFalse(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);
      
      
      // Test no election policy, was not old master, is not new master
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.isMasterReplica(SERVICE_HA_NAME)).andReturn(false);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertFalse(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);
      
      
      // Test no election policy, was not old master, is new master
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.isMasterReplica(SERVICE_HA_NAME)).andReturn(true);
      
      this.partition.callAsynchMethodOnCluster(EasyMock.same(SERVICE_HA_NAME), EasyMock.eq("stopOldMaster"), EasyMock.aryEq(new Object[0]), EasyMock.aryEq(new Class[0]), EasyMock.eq(true));
      this.lifecycle.startSingleton();
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);

      
      // Test no election policy, was old master, is new master, restart on merge, not merged
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.isMasterReplica(SERVICE_HA_NAME)).andReturn(true);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, false);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);

      
      // Test no election policy, was old master, is new master, restart on merge, merged
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.isMasterReplica(SERVICE_HA_NAME)).andReturn(true);

      this.lifecycle.stopSingleton();
      this.lifecycle.startSingleton();
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);

      
      // Test no election policy, was old master, is new master, no restart on merge
      
      this.singleton.setRestartOnMerge(false);
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.isMasterReplica(SERVICE_HA_NAME)).andReturn(true);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);
      
      
      // Test no election policy, was old master, not new master
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.isMasterReplica(SERVICE_HA_NAME)).andReturn(false);

      this.lifecycle.stopSingleton();
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, false);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertFalse(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);
      
      // Set election policy
      this.singleton.setElectionPolicy(this.electionPolicy);

      
      // Test null replicants, was not old master, is not new master
      
      List<ClusterNode> nodes = null;
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertFalse(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);
      
      
      // Test empty replicants, was not old master, is not new master
      
      nodes = Collections.emptyList();
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertFalse(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);
      
      
      // Test one replicant, was not old master, is not new master
      
      ClusterNode ourNode = EasyMock.createMock("ours", ClusterNode.class);
      ClusterNode otherNode = EasyMock.createMock("other", ClusterNode.class);
      
      nodes = Collections.singletonList(otherNode);
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      EasyMock.expect(this.partition.getClusterNode()).andReturn(ourNode);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle);
      
      assertFalse(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle);
      
      
      // Test many replicants, was not old master, is not new master
      nodes = Arrays.asList(ourNode, otherNode);
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      EasyMock.expect(this.electionPolicy.elect(nodes)).andReturn(otherNode);
      EasyMock.expect(this.partition.getClusterNode()).andReturn(ourNode);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      assertFalse(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      
      // Test many replicants, was not old master, is new master
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      EasyMock.expect(this.electionPolicy.elect(nodes)).andReturn(ourNode);
      EasyMock.expect(this.partition.getClusterNode()).andReturn(ourNode);
      
      this.partition.callAsynchMethodOnCluster(EasyMock.same(SERVICE_HA_NAME), EasyMock.eq("stopOldMaster"), EasyMock.aryEq(new Object[0]), EasyMock.aryEq(new Class[0]), EasyMock.eq(true));
      this.lifecycle.startSingleton();
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle, this.electionPolicy);

      
      // Test many replicants, was old master, is new master, no restart on merge
      
      this.singleton.setRestartOnMerge(false);
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      EasyMock.expect(this.electionPolicy.elect(nodes)).andReturn(ourNode);
      EasyMock.expect(this.partition.getClusterNode()).andReturn(ourNode);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      
      // Test one replicant, was old master, is new master, no restart on merge
      
      nodes = Collections.singletonList(ourNode);
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      EasyMock.expect(this.partition.getClusterNode()).andReturn(ourNode);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      
      // Test many replicants, was old master, is new master, restart on merge, not merged
      
      nodes = Arrays.asList(ourNode, otherNode);
      
      this.singleton.setRestartOnMerge(true);

      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      EasyMock.expect(this.electionPolicy.elect(nodes)).andReturn(ourNode);
      EasyMock.expect(this.partition.getClusterNode()).andReturn(ourNode);
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, false);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle, this.electionPolicy);

      
      // Test many replicants, was old master, is new master, restart on merge, merged
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      EasyMock.expect(this.electionPolicy.elect(nodes)).andReturn(ourNode);
      EasyMock.expect(this.partition.getClusterNode()).andReturn(ourNode);

      this.lifecycle.stopSingleton();
      this.lifecycle.startSingleton();
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      assertTrue(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle, this.electionPolicy);

      
      // Test many replicants, was old master, not new master
      
      EasyMock.expect(this.partition.getDistributedReplicantManager()).andReturn(this.drm);
      EasyMock.expect(this.drm.lookupReplicantsNodes(SERVICE_HA_NAME)).andReturn(nodes);
      EasyMock.expect(this.electionPolicy.elect(nodes)).andReturn(otherNode);
      EasyMock.expect(this.partition.getClusterNode()).andReturn(ourNode);

      this.lifecycle.stopSingleton();
      
      EasyMock.replay(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      this.singleton.replicantsChanged(SERVICE_HA_NAME, list, 101, true);
      
      EasyMock.verify(this.partition, this.drm, this.lifecycle, this.electionPolicy);
      
      assertFalse(this.singleton.isMasterNode());
      
      EasyMock.reset(this.partition, this.drm, this.lifecycle, this.electionPolicy);
   }
}
